<?php
namespace KnitPay\Gateways\PayU;

use Pronamic\WordPress\Pay\Core\Gateway as Core_Gateway;
use Pronamic\WordPress\Pay\Payments\Payment;
use Pronamic\WordPress\Pay\Payments\PaymentStatus as Core_Statuses;
use Exception;
use WP_Error;

/**
 * Title: PayU Gateway
 * Copyright: 2020-2021 Knit Pay
 *
 * @author Knit Pay
 * @version 5.4.0
 * @since 5.4.0
 */
class Gateway extends Core_Gateway {

	/**
	 * Client.
	 *
	 * @var Client
	 */
	private $client;

	/**
	 * Constructs and initializes an PayU gateway
	 *
	 * @param Config $config
	 *            Config.
	 */
	public function __construct( Config $config ) {
		parent::__construct( $config );

		$this->set_method( self::METHOD_HTML_FORM );

		// Supported features.
		 $this->supports = array(
			 'payment_status_request',
		 );

		 // Client.
		 $this->client = new Client( $config, self::MODE_TEST === $config->mode );
	}

	/**
	 * Get supported payment methods
	 *
	 * @see Core_Gateway::get_supported_payment_methods()
	 */
	public function get_supported_payment_methods() {
		return array(
			PaymentMethods::CREDIT_CARD,
			PaymentMethods::PAY_U,
		);
	}

	/**
	 * Get available payment methods.
	 *
	 * @return array<int, string>
	 * @see Core_Gateway::get_available_payment_methods()
	 */
	public function get_available_payment_methods() {
		return $this->get_supported_payment_methods();
	}

	/**
	 * Start.
	 *
	 * @see Core_Gateway::start()
	 *
	 * @param Payment $payment
	 *            Payment.
	 */
	public function start( Payment $payment ) {
		$payment_currency = $payment->get_total_amount()
			->get_currency()
			->get_alphabetic_code();
		if ( isset( $payment_currency ) && 'INR' !== $payment_currency ) {
			$currency_error = 'PayU only accepts payments in Indian Rupees. If you are a store owner, kindly activate INR currency for ' . $payment->get_source() . ' plugin.';
			$this->error    = new WP_Error( 'payu_error', $currency_error );
			return;
		}

		$payment->set_transaction_id( $payment->key . '_' . $payment->get_id() );
		$payment->set_action_url( $this->client->get_payment_server_url() . '/_payment' );
	}

	/**
	 * Update status of the specified payment.
	 *
	 * @param Payment $payment
	 *            Payment.
	 */
	public function update_status( Payment $payment ) {
		if ( Core_Statuses::SUCCESS === $payment->get_status() ) {
			return;
		}

		try {
			$transaction = $this->client->verify_payment( $payment->get_transaction_id() );
			if ( $transaction->txnid !== $payment->get_transaction_id() ) {
				$payment->add_note( 'Something went wrong: ' . print_r( $transaction, true ) );
				return;
			}

			if ( isset( $transaction->status ) ) {
				$transaction_status = $transaction->status;

				if ( Statuses::SUCCESS === $transaction_status ) {
					$payment->set_transaction_id( $transaction->mihpayid );
				}

				$note = '<strong>PayU Transaction Details:</strong><br>' . print_r( $transaction, true );

				$payment->set_status( Statuses::transform( $transaction_status ) );
				$payment->add_note( $note );
			}
		} catch ( Exception $e ) {
			$this->error = new WP_Error( 'payu_error', $e->getMessage() );
		}

	}

	/**
	 * Redirect via HTML.
	 *
	 * @see Core_Gateway::get_output_fields()
	 *
	 * @param Payment $payment
	 *            Payment.
	 */
	public function get_output_fields( Payment $payment ) {

		$merchant_key  = $this->config->merchant_key;
		$merchant_salt = $this->config->merchant_salt;

		$txnid        = $payment->get_transaction_id();
		$amount       = $payment->get_total_amount()->number_format( null, '.', '' );
		$product_info = $payment->get_description();

		$first_name = '';
		$last_name  = '';
		$customer   = $payment->get_customer();
		if ( null !== $customer->get_name() ) {
			$first_name = $customer->get_name()->get_first_name();
			$last_name  = $customer->get_name()->get_last_name();
		}
		$email = $customer->get_email();

		$phone     = '';
		$address   = '';
		$address_2 = '';
		$city      = '';
		$state     = '';
		$country   = '';
		$zipcode   = '';

		$billing_address = $payment->get_billing_address();
		if ( null !== $billing_address ) {
			if ( ! empty( $billing_address->get_phone() ) ) {
				$phone = $billing_address->get_phone();
			}
			$address   = $billing_address->get_line_1();
			$address_2 = $billing_address->get_line_2();
			$city      = $billing_address->get_city();
			$state     = $billing_address->get_region();
			$country   = $billing_address->get_country();
			$zipcode   = $billing_address->get_postal_code();
		}

		$udf1 = null;
		$udf2 = null;
		$udf3 = null;
		$udf4 = null;
		$udf5 = null;

		$str = "{$merchant_key}|{$txnid}|{$amount}|{$product_info}|{$first_name}|{$email}|{$udf1}|{$udf2}|{$udf3}|{$udf4}|{$udf5}||||||{$merchant_salt}";

		$hash = strtolower( hash( 'sha512', $str ) );

		$return_url = $payment->get_return_url();

		return array(
			'key'         => $merchant_key,
			'txnid'       => $txnid,
			'amount'      => $amount,
			'productinfo' => $product_info,
			'firstname'   => $first_name,
			'lastname'    => $last_name,
			'address1'    => $address,
			'address2'    => $address_2,
			'city'        => $city,
			'state'       => $state,
			'country'     => $country,
			'zipcode'     => $zipcode,
			'email'       => $email,
			'phone'       => $phone,
			'surl'        => $return_url,
			'furl'        => $return_url,
			'hash'        => $hash,
			'udf5'        => $udf5,
			// 'pg'               => 'DC' //TODO @see https://devguide.payu.in/docs/developers-guide/checkout/payu-prebuilt-checkout-overview/prebuilt-checkout-integration/
		);
	}
}
